-- alu_segment.vhd
--
-- The entity here implements an n-bit wide slice of the ALU.
-- The width is set by the `bits' generic parameter.
--

library ieee ;
use ieee . std_logic_1164 . all ;
use ieee . std_logic_arith . all ;
use ieee . std_logic_unsigned . all ;
use m68k_types . all ;



entity alu_segment is
    generic ( bits      : integer ) ;
    port (  input_b     : in std_logic_vector (( bits - 1 ) downto 0 ) ;
            input_a     : in std_logic_vector (( bits - 1 ) downto 0 ) ;
            output      : out std_logic_vector (( bits - 1 ) downto 0 ) ;
            carry_in    : in std_logic ;
            carry_out   : out std_logic ;
            alu_internal_op
                        : in alu_internal_op_type ) ;
end entity alu_segment ;

architecture basic of alu_segment is
begin

    -- ALU internal operation codes
    -- ALU_INT_OR       output <- A or B
    -- ALU_INT_AND      output <- A and B
    -- ALU_INT_SUB      output <- A - B
    -- ALU_INT_ADD      output <- A + B
    -- ALU_INT_EOR      output <- A xor B
    -- ALU_INT_CMP      output <- B - A
    -- ALU_INT_REV_SUB  output <- B - A
    -- ALU_INT_REV_CMP  output <- B - A


    process ( input_b , input_a , alu_internal_op , carry_in ) is
        variable int_input_b    : std_logic_vector 
                                    (( bits + 1 ) downto 0 ) ;
        variable int_input_a    : std_logic_vector 
                                    (( bits + 1 ) downto 0 ) ;
        variable int_output     : std_logic_vector 
                                    (( bits + 1 ) downto 0 ) ;
    begin
        int_input_a ( bits + 1 ) := '0' ;
        int_input_b ( bits + 1 ) := '0' ;
        
        if ( alu_internal_op = ALU_INT_REV_SUB )
        or ( alu_internal_op = ALU_INT_REV_CMP )
        then
            -- Reverse subtraction. The result is B - A.
            -- 2's complement of A is obtained:
            int_input_b ( 0 ) := carry_in ;
            int_input_b ( bits downto 1 ) := input_b ;

            int_input_a ( 0 ) := '1' ;
            int_input_a ( bits downto 1 ) := not input_a ;
        else
            int_input_a ( 0 ) := carry_in ;
            int_input_a ( bits downto 1 ) := input_a ;

            int_input_b ( 0 ) := '1' ;
            if ( alu_internal_op = ALU_INT_SUB )
            or ( alu_internal_op = ALU_INT_CMP )
            then
                -- For subtraction operations, 2's complement of B
                -- is obtained.
                int_input_b ( bits downto 1 ) := not input_b ;
            else
                int_input_b ( bits downto 1 ) := input_b ;
            end if ;
        end if ;

        case alu_internal_op is
        when ALU_INT_AND =>
                int_output := ( int_input_a and int_input_b ) ;
        when ALU_INT_EOR =>
                int_output := ( int_input_a xor int_input_b ) ;
        when ALU_INT_OR =>
                int_output := ( int_input_a or int_input_b ) ;
        when others => -- ADD, SUB, CMP
                int_output := ( int_input_a + int_input_b ) ;
        end case ;

        carry_out <= int_output ( bits + 1 ) ;
        output <= int_output ( bits downto 1 ) ;
    end process ;

end architecture basic ;

    
