
    rom : entity program_rom
        port map ( address => memory_address ,
                    data => rom_output ,
                    clock => clock ) ;      -- memory access in +ve cycle

    ram : entity dp_ram
        generic map ( address_width => 12 , -- 4K RAM
                    data_width => 8 )
        port map ( clock => clock ,         -- memory access in +ve cycle
                    write => ram_write_enable ,
                    address1 => memory_address ( 11 downto 0 ) ,
                    address2 ( 3 downto 0 ) => switches ( 3 downto 0 ) ,
                    address2 ( 11 downto 4 ) => "00000000" ,
                    data_in => memory_input ,
                    data_out1 => ram_output ,
                    data_out2 => debug_memory_out ) ;

    -- Memory map:
    -- 0000-0fff    ROM
    -- 1000-1fff    RAM
    -- 8000         Output
    
    memory_map_process  : process ( rom_output , 
                ram_output , memory_address , clock ,
                memory_write_enable , memory_input ) is
    begin
        case memory_address ( 15 downto 12 ) is
        when "0000" =>  memory_output <= rom_output ;
                        ram_write_enable <= '0' ;

        when "0001" =>  memory_output <= ram_output ;
                        ram_write_enable <= memory_write_enable ;

        when "1000" =>  memory_output <= ( others => '0' ) ;
                        if ( memory_write_enable = '1' )
                        and ( memory_address ( 11 downto 0 ) =
                                conv_std_logic_vector ( 0 , 12 ))
                        then
                            -- memory_input contains the
                            -- data to be output.
                            if ( clock = '1' )
                            and ( clock'event )
                            then
                                last_output <= memory_input ;
                            end if ;
                        end if ;
                        ram_write_enable <= '0' ;

        when others =>  memory_output <= ( others => '0' ) ;
                        ram_write_enable <= '0' ;
        end case ;
    end process memory_map_process ;

    memory_address_mux : process ( mar_source ,
            pc_register , operand_address ) is
    begin
        case mar_source is
        when PC_TO_MAR =>
                memory_address <= pc_register ;
        when others => -- OA_TO_MAR
                memory_address <= operand_address ;
        end case ;
    end process memory_address_mux ;

    memory_output_latch : process ( clock , memory_output ) is
    begin
        if ( clock = '0' )
        and ( clock'event )
        then
            last_memory_output <= memory_output ;
        end if ;
    end process memory_output_latch ;

    memory_input_mux : process ( mdr_source ,
            operand_value ) is
    begin
        case mdr_source is
        when OV_0_TO_MDR =>
                    memory_input <= operand_value ( 7 downto 0 ) ;
                    memory_write_enable <= '1' ;
        when OV_1_TO_MDR =>
                    memory_input <= operand_value ( 15 downto 8 ) ;
                    memory_write_enable <= '1' ;
        when OV_2_TO_MDR =>
                    memory_input <= operand_value ( 23 downto 16 ) ;
                    memory_write_enable <= '1' ;
        when OV_3_TO_MDR =>
                    memory_input <= operand_value ( 31 downto 24 ) ;
                    memory_write_enable <= '1' ;
        when others =>
                    -- Writing is turned off, so the source of memory_input
                    -- is unimportant. Choose something that has been used
                    -- before to simplify this mux.
                    memory_input <= operand_value ( 31 downto 24 ) ;
                    memory_write_enable <= '0' ;
        end case ;
    end process memory_input_mux ;

    register_transfers : process ( ir_source , 
            clock , last_memory_output , operand_value_source ,
            alu_output , immediate_data_source , pc_source ,
            operand_address , operand_address_source ) is
    begin
        -- The data that is transfered from memory MUST have been
        -- fetched on the previous clock cycle. If it was fetched
        -- earlier or later than that, it won't be available.
        if ( clock = '1' )
        and ( clock'event )
        then
            case ir_source is
            when MDR_TO_IR_1 =>
                        instruction_register ( 15 downto 8 ) 
                                        <= last_memory_output ;
            when MDR_TO_IR_0 =>
                        instruction_register ( 7 downto 0 ) 
                                        <= last_memory_output ;
            when others => null ;
            end case ;

            case operand_value_source is
            when MDR_TO_OV_3 =>
                        operand_value ( 31 downto 24 ) 
                                        <= last_memory_output ;
            when MDR_TO_OV_2 =>
                        operand_value ( 23 downto 16 ) 
                                        <= last_memory_output ;
            when MDR_TO_OV_1 =>
                        operand_value ( 15 downto 8 ) 
                                        <= last_memory_output ;
            when MDR_TO_OV_0 =>
                        operand_value ( 7 downto 0 ) 
                                        <= last_memory_output ;
            when ALU_TO_OV =>
                        operand_value <= alu_output ;
            when others => null ;
            end case ;

            case immediate_data_source is
            when MDR_TO_IDR_3 =>
                        immediate_data_reg ( 31 downto 24 ) 
                                        <= last_memory_output ;
            when MDR_TO_IDR_2 =>
                        immediate_data_reg ( 23 downto 16 ) 
                                        <= last_memory_output ;
            when MDR_TO_IDR_1 =>
                        immediate_data_reg ( 15 downto 8 ) 
                                        <= last_memory_output ;
            when MDR_TO_IDR_0 =>
                        immediate_data_reg ( 7 downto 0 ) 
                                        <= last_memory_output ;
            when MDR_TO_IDR_1_SE =>
                        immediate_data_reg ( 31 downto 16 ) 
                            <= ( others => last_memory_output ( 7 ) ) ;
                        immediate_data_reg ( 15 downto 8 ) 
                                        <= last_memory_output ;
            when MDR_TO_IDR_0_SE =>
                        immediate_data_reg ( 31 downto 8 ) 
                            <= ( others => last_memory_output ( 7 ) ) ;
                        immediate_data_reg ( 7 downto 0 ) 
                                        <= last_memory_output ;
            when others => null ;
            end case ;

            case pc_source is
            when ALU_TO_PC =>
                        pc_register <= alu_output ;
            when others => null ;
            end case ;

            case operand_address_source is
            when ALU_TO_OA =>
                        operand_address <= alu_output ;
            when MDR_TO_OA_3 =>
                        operand_address ( 31 downto 24 ) <= last_memory_output ;
            when MDR_TO_OA_2 =>
                        operand_address ( 23 downto 16 ) <= last_memory_output ;
            when MDR_TO_OA_1_SE =>
                        operand_address ( 31 downto 16 ) <= 
                                ( others => last_memory_output ( 7 )) ;
                        operand_address ( 15 downto 8 ) <= last_memory_output ;
            when MDR_TO_OA_1 =>
                        operand_address ( 15 downto 8 ) <= last_memory_output ;
            when MDR_TO_OA_0 =>
                        operand_address ( 7 downto 0 ) <= last_memory_output ;
            when others => null ;
            end case ;
        end if ;
    end process register_transfers ;




