#include <stdio.h>

#include <sys/types.h>
#include <regex.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "opcode_database.h"
#include "utils.h"




/****** Opcode_Database public methods ******/

/** Opcode_Database constructor
 *
 * Create a new Opcode_Database object. Only one of these will be needed
 * per run.
 */
Opcode_Database :: Opcode_Database () : Opcode_Map_Reader ()
{
    finalised = false ;
}


/** Require_Opcode
 *
 * Notifies the database that the given opcode will be required by the program.
 * The opcode micro subroutine name is returned.
 */
const char * Opcode_Database :: Require_Opcode ( unsigned opcode )
{
    assert ( ! finalised ) ;

    MESSAGE2 ( "Adding requirement for opcode 0x%04x.. " , opcode ) ;

    Opcode_Family   * family_for_op = 
            (Opcode_Family *) ndfa . Enable_Accept_State ( opcode ) ;

    if ( family_for_op == 0L )
    {
        throw new UnavailableOpcodeException () ;
    }
    MESSAGE2 ( "%s\n" , family_for_op -> Get_Micro_Sub_Name () ) ;

    return family_for_op -> Get_Micro_Sub_Name () ;
}


/** Finalise_DFA
 *
 * Finalises the database, making it ready for VHDL generation.
 * After this, no changes can be made.
 *
 * The state machine passed in must be a finalised master state machine with
 * _every_ required microsubroutine already present - i.e. one built
 * by State_Machine_Loader::Build_Master_State_Machine().
 */
void Opcode_Database :: Finalise_DFA ( State_Machine * sm )
{
    assert ( ! finalised ) ;

    MESSAGE ( "Finalising instruction decode DFA..\n" ) ;
    MESSAGE2 ( "Dead branch elimination.. (size before = %d)\n" ,
                                ndfa . Get_Size () ) ;

    ndfa . Delete_Dead_Branches () ;

    MESSAGE2 ( "Compression.. (size before = %d)\n" ,
                                ndfa . Get_Size () ) ;

    ndfa . Compress () ;

    MESSAGE2 ( "Finished. Size = %d.\n" , ndfa . Get_Size () ) ;

    assert ( sm -> Is_Finalised () ) ;

    if ( g_verbose_setting >= VERBOSE_VERY_HIGH )
    {
        ndfa . Print_NDFA_DAG ( true , stdout ) ;
    }
        

    MESSAGE2 ( "Linking subroutines..\n" ) ;

    Link_Subroutines ( sm ) ;

    MESSAGE2 ( "Done.\n" ) ;

    finalised = true ;
}



/** New_Accept_State
 *
 * Create a new accept state/reuse an existing accept state. This method
 * is called from inherited method Read_Opcode_Map as data is read from the
 * file.
 */
Accept_State * Opcode_Database :: New_Accept_State ( 
                const char * cmp_bit_pattern , 
                const char * opcode_name , 
                const char * dea , 
                const char * micro_sub_name , 
                const char * optimisation_info , 
                const char * comment )
{
    Opcode_Family *                     opcode_family ;
    pair<Opcode_Set_Iterator, bool>     insert_result ;


    /* Look at the micro subroutine name. Is this opcode decoded
     * by an existing micro subroutine? We try to create a new
     * one.. */
    opcode_family = new Opcode_Family ( micro_sub_name ) ;

    /* and try inserting it into the opcode list */
    insert_result = opcode_set . insert ( opcode_family ) ;

    if ( ! insert_result . second )
    {
        MESSAGE3 ( "  (opcode family already present)\n" ) ;
        /* it wasn't correctly inserted. Must already be in there. */
        delete opcode_family ;

        opcode_family = (* ( insert_result . first )) ;
    } else {
        MESSAGE3 ( "  (opcode family added)\n" ) ;
    }

    opcode_family -> Add_Opcode ( opcode_name , comment ) ;

    return opcode_family ;
}



/** List_Required_Micro_Subroutines
 *
 * A Label_Set list is produced, containing the set of microsubroutines
 * required to run the program. All required opcodes should have been
 * added to the database. */
Label_Set Opcode_Database :: List_Required_Micro_Subroutines ()
{
    Opcode_Set_Iterator     iter = opcode_set . begin () ;
    Label_Set               msub_list ;

    while ( iter != opcode_set . end () )
    {
        Opcode_Family * item = (* iter) ;

        if ( item -> Is_Enabled ())
        {
            msub_list . insert ( item -> Get_Micro_Sub_Name () ) ;
        }
        iter ++ ;
    }
    return msub_list ;
}


/****** Opcode_Database private methods ******/


/** Link_Subroutines
 *
 * All microsubroutines that are required to run the program are linked
 * to their starting state in the given state machine. */
void Opcode_Database :: Link_Subroutines ( State_Machine * sm )
{
    Opcode_Set_Iterator     iter = opcode_set . begin () ;

    while ( iter != opcode_set . end () )
    {
        Opcode_Family * item = (* iter) ;

        if ( item -> Is_Enabled ())
        {
            item -> Link_Subroutine ( sm ) ;
        }
        iter ++ ;
    }
}

