

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <limits.h>

#include <map>
#include <list>
#include <iostream>

#include "utils.h"
#include "state.h"



/****** State public methods ******/

/** State destructor
 *
 * Frees the state and associated command list.
 */
State :: ~State ()
{
    /* Free the command list */
    Command_List_Iterator   iter ;
    Command *               item ;
    
    iter = commands . begin () ;
    while ( iter != commands . end () )
    {
        item = (* iter) ;
        delete item ;
        iter ++ ;
    }

    if ( state_name != 0L )
    {
        delete [] state_name ;
    }
}

/** Add_Command
 *
 * Add an extra command to the end of the list of VHDL 
 * commands for the given state. */
void State :: Add_Command ( Command * cmd )
{
    commands . push_back ( cmd ) ;
}


/** Depends_On
 *
 * Scan the command list for CALLs/JUMPs and return their labels as a set */
Label_Set State :: Depends_On ()
{
    Label_Set               depends_on_set ;
    Command_List_Iterator   iter ;
    Command *               item ;
    
    iter = commands . begin () ;
    while ( iter != commands . end () )
    {
        item = (* iter) ;

        if (( item -> Get_Type () == Command :: JUMP )
        || ( item -> Get_Type () == Command :: CALL ))
        {
            depends_on_set . insert ( item -> Get_Data () ) ;
        }
        iter ++ ;
    }
    return depends_on_set ;
}


/** Generate_VHDL
 *
 * VHDL is generated for the state.
 */
void State :: Generate_VHDL ( FILE * output , State_Map * definitions )
{
    Command_List_Iterator   command_list_pos ;

    fprintf ( output , "\nwhen " ) ;
    Print_Binary ( output , abs_state_number , width_of_state_number ) ;
    fprintf ( output , " => -- 0x%04x (%d) %s\n" ,
                abs_state_number , abs_state_number , 
                state_name == 0L ? "none" : state_name ) ;

    command_list_pos = commands . begin () ;
    while ( command_list_pos != commands . end () ) 
    {
        Command *           current_command = (* command_list_pos) ;
        State *             target_mc ;
        
        switch ( current_command -> Get_Type ())
        {
            case Command :: VHDL :
                fputs ( current_command -> Get_Data () , output ) ;
                fputs ( "\n" , output ) ;
                break ;
            case Command :: IDECODE :
                /* Instruction decode. This is much like a CALL,
                 * except the address to be called comes from the
                 * instruction decoding logic. */
                fprintf ( output , "\n\t-- Instruction decode.\n"
                        "\tcall_state <= decoded_state ;\n"
                        "\tcall_requested <= '1' ;\n"
                        "\treturn_requested <= '0' ;\n" ) ;
                break ;
            case Command :: JUMP :
            case Command :: CALL :
                /* The target machine should be part of this machine
                 * already if the dependencies have been properly
                 * satisfied. If it is not, we stop now. */
                assert ( definitions -> count 
                                ( current_command -> Get_Data () ) != 0 ) ;

                target_mc = (* definitions) [ current_command -> Get_Data () ] ;

                assert ( target_mc != 0 ) ;

                fprintf ( output ,
                    "\n\t-- %s %s:\n" 
                    "\tcall_state <= " ,
                    ( current_command -> Get_Type () == Command :: CALL ) ? 
                                                "CALL" : "JUMP" ,
                    current_command -> Get_Data () ) ;

                Print_Binary ( output ,
                    target_mc -> abs_state_number , /* target state */
                    width_of_state_number ) ;

                fprintf ( output , " ;\n"
                        "\tcall_requested <= '1' ;\n"
                        "\treturn_requested <= '%d' ;\n" ,
                        ( current_command -> Get_Type () == Command :: CALL ) ?
                                                0 : 1 ) ;
                break ;
            case Command :: RETURN :
                fprintf ( output , 
                        "\treturn_requested <= '1' ;\n" ) ;
                break ;
            default :
                assert ( 0 ) ;
        }
        command_list_pos ++ ;
    }
}

/** Generate_Link_To_State_VHDL
 *
 * Instruction decoder VHDL is generated to cause a jump to the state.
 */
void State :: Generate_Link_To_State_VHDL ( FILE * output , 
                                            const char * indent )
{
    fprintf ( output , "%sdecoded_state <= " , indent ) ;

    Print_Binary ( output , abs_state_number ,
                width_of_state_number ) ;

    fprintf ( output , " ; -- (%s)\n" ,
                state_name == 0L ? "none" : state_name ) ;
}

/** Set_State_Name
 *
 * Changes the name of the state. If the name is already known,
 * the space is freed before it is changed.
 */
void State :: Set_State_Name ( const char * name )
{
    if ( state_name != 0L )
    {
        delete [] state_name ;
    }
    state_name = Copy_String ( name ) ;
}


/****** Command public methods ******/

/** Command constructor
 *
 * Creates a new Command class.
 */
Command :: Command ( Command_Type t , const char * cmd_data )
{
    type = t ;
    if ( cmd_data == 0L )
    {
        data = 0L ;
    } else {
        data = Copy_String ( cmd_data ) ;
    }
}

/** Command destructor
 *
 * Deletes a Command class.
 */
Command :: ~Command ()
{
    /* Free the data associated with the command, if any */
    if ( data != 0L )
    {
        delete [] data ;
    }
}



