#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <limits.h>

#include <sys/types.h>
#include <dirent.h>
#include <fnmatch.h>

#include <map>
#include <list>
#include <iostream>

#include "utils.h"
#include "state_machine_loader.h"


/****** State_Machine public methods ******/

/** State_Machine_Loader destructor
 *
 * All state machines are deleted.
 */
State_Machine_Loader :: ~State_Machine_Loader ()
{
    State_Machine_Set_Iterator iter = machines . begin () ;

    while ( iter != machines . end () )
    {
        State_Machine * item = (* iter) ;

        delete item ;

        iter ++ ;
    }
}

/** Add_State_Machine
 *
 * A state machine is loaded from the given file, and added to the loader's
 * internal database.
 */
void State_Machine_Loader :: Add_State_Machine ( const char * filename )
{
    assert ( ! is_finalised ) ;

    MESSAGE2 ( "Loading state machine file %s\n" , filename ) ;

    State_Machine * sm = new State_Machine ( filename ) ;

    /* The new state machine is added to the list of machines */
    machines . insert ( sm ) ;

    /* The provides list of the machine is examined. */
    Label_Set           provides_list = sm -> Provides () ;
    Label_Set_Iterator  iter = provides_list . begin () ;

    while ( iter != provides_list . end () )
    {
        const char * state_name = (* iter) ;

        if ( provides_map . count ( string ( state_name ) ) > 0 )
        {
            throw new Duplicate_Label_Exception ( state_name ) ;
        }
        provides_map [ string ( state_name ) ] = sm ;

        iter ++ ;
    }
}

/** Add_State_Machine_Directory
 *
 * All state machine (.sm) files are loaded from the given directory.
 */
void State_Machine_Loader :: Add_State_Machine_Directory ( const char * dir )
{
    assert ( ! is_finalised ) ;

    DIR *           dird = opendir ( dir ) ;
    struct dirent * entry ;

    if ( dird == 0L )
    {
        throw new Dir_Access_Exception ( dir ) ;
    }
    MESSAGE ( "Scanning state machine directory %s\n" , dir ) ;


    while (( entry = readdir ( dird ) ) != NULL )
    {
        char * filename = new char [ strlen ( entry -> d_name )
                                            + strlen ( dir ) + 2 ] ;
        strcpy ( filename , dir ) ;
        strcat ( filename , "/" ) ;
        strcat ( filename , entry -> d_name ) ;

        if ( fnmatch ( STATE_MACHINE_GLOB , filename , 0 ) == 0 )
        {
            Add_State_Machine ( filename ) ;
        }

        delete [] filename ;
    }
    closedir ( dird ) ;
}


/** Require_Microsub
 * 
 * Indicate to the database that a particular microsubroutine is required.
 */
void State_Machine_Loader :: Require_Microsub ( const char * sub_name )
{
    assert ( ! is_finalised ) ;

    /* The machine containing this subroutine will certainly be required. */
    MESSAGE2 ( "Adding microsubroutine %s.." , sub_name ) ;
    State_Machine * sm = Get_State_Machine_For_Name ( sub_name ) ;

    /* Is sm already on the required list of machines? */
    if ( required_machines . count ( sm ) == 0 )
    {
        /* No... add it now. */
        required_machines . insert ( sm ) ;

        MESSAGE2 ( "added\n" ) ;
        /* Also add everything it depends on */
        Require_Microsubs ( sm -> Depends_On () ) ;
    } else {
        MESSAGE2 ( "already present\n" ) ;
    }
}


/** Require_Microsubs
 * 
 * Indicate to the database that all the members of a set of
 * microsubroutines are required.
 */
void State_Machine_Loader :: Require_Microsubs ( Label_Set list )
{
    assert ( ! is_finalised ) ;

    Label_Set_Iterator  iter = list . begin () ;

    while ( iter != list . end () )
    {
        const char * state_name = (* iter) ;
        Require_Microsub ( state_name ) ;

        iter ++ ;
    }
}

/** Build_Master_Machine
 *
 * Creates a new state machine containing all the machines required
 * to execute the program. 
 *
 * Once called, no more changes can be made to the state machine loader
 * database (they could invalidate the machine).
 */
State_Machine * State_Machine_Loader :: Build_Master_Machine ( 
                                        const char * root_sub_name )
{
    if ( is_finalised )
    {
        /* machine already built */
        return master_machine ;
    }

    MESSAGE ( "Building master state machine..\n" ) ;

    /* The master machine is based on the machine containing the
     * root subroutine named here.
     * Ensure that all dependencies are met, and that the
     * named subroutine exists. */
    Require_Microsub ( root_sub_name ) ;
        
    master_machine = Get_State_Machine_For_Name ( root_sub_name ) ;

    /* Add all other required machines to the master machine. This
     * will effectively concatenate them on the end. */
    State_Machine_Set      required_machines_copy = required_machines ;
    State_Machine_Set_Iterator iter ;
    
    required_machines_copy . erase ( master_machine ) ;

    iter = required_machines_copy . begin () ;
    while ( iter != required_machines_copy . end () )
    {
        State_Machine * item = (* iter) ;

        master_machine -> Incorporate_Sub_Machine ( item ) ;

        iter ++ ;
    }

    is_finalised = true ; /* No more changes can be made to the database */

    master_machine -> Finalise_SM () ;

    /* Debugging information */
    MESSAGE2 ( "State machine report:\n" ) ;
    iter = machines . begin () ;
    while ( iter != machines . end () )
    {
        State_Machine * item = (* iter) ;

        if ( required_machines . count ( item ) > 0 )
        {
            MESSAGE2 ( "includes %s\n" , item -> Get_Name () ) ;
        } else {
            MESSAGE2 ( "   omits %s\n" , item -> Get_Name () ) ;
        }
        iter ++ ;
    }
    MESSAGE2 ( "Total: %d machines of %d included.\n" ,
            required_machines . size () , machines . size () ) ;

    return master_machine ;
}

/** Get_State_Machine_For_Name
 *
 * Finds the state machine which provides the given state/microsubroutine.
 */
State_Machine * State_Machine_Loader :: 
                    Get_State_Machine_For_Name ( const char * sub_name )
{
    if ( provides_map . count ( string ( sub_name ) ) == 0 )
    {
        /* This subroutine doesn't seem to exist! */
        throw new Unavailable_Microsub_Exception ( sub_name ) ;
    }
    return provides_map [ string ( sub_name ) ] ;
}



